from ewah.constants import EWAHConstants as EC
from ewah.hooks.base import EWAHBaseHook

import snowflake.connector
from typing import Optional, List, Union


class EWAHSnowflakeHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "user": "login",
        "account": "host",
    }

    conn_name_attr = "ewah_snowflake_conn_id"
    default_conn_name = "ewah_snowflake_default"
    conn_type = "ewah_snowflake"
    hook_name = "EWAH Snowflake Connection"

    def __init__(self, *args, database: Optional[str] = None, **kwargs):
        self.database = database
        return super().__init__(*args, **kwargs)

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "port", "schema"],
            "relabeling": {
                "password": "Password",
                "login": "User",
                "host": "Account",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from wtforms import StringField

        return {
            "extra__ewah_snowflake__database": StringField(
                "Default Database",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_snowflake__role": StringField(
                "Role",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_snowflake__warehouse": StringField(
                "Warehouse",
                widget=BS3TextFieldWidget(),
            ),
        }

    @property
    def snow_conn(self):
        if not hasattr(self, "_snow_conn"):
            self._snow_conn = snowflake.connector.connect(
                user=self.conn.user,
                password=self.conn.password,
                role=self.conn.role,
                warehouse=self.conn.warehouse,
                account=self.conn.account,
                database=self.database or self.conn.database,
            )
        return self._snow_conn

    @staticmethod
    def _adjust_sql(sql):
        sql = sql.strip()
        if sql[-1:] == ";":
            sql = sql[:-1].strip()
        return sql

    @property
    def cursor(self):
        if not hasattr(self, "_cur"):
            self._cur = self.snow_conn.cursor()
            # Start a transaction!
            self._cur.execute("BEGIN;")
        return self._cur

    def commit(self):
        self.cursor.execute("COMMIT;")
        self.cursor.execute("BEGIN;")

    def rollback(self):
        self.cursor.execute("ROLLBACK;")
        self.cursor.execute("BEGIN;")

    def execute(
        self, sql: str, params: Optional[dict] = None, commit: bool = False, cursor=None
    ) -> None:
        self.log.info(
            "Executing SQL:\n\n{0}\n\nWith params:\n{1}".format(
                sql,
                "\n".join(
                    [
                        "{0}: {1}".format(key, str(value))
                        for (key, value) in params.items()
                    ]
                )
                if params
                else "No params!",
            )
        )
        sql = self._adjust_sql(sql)
        for statement in sql.split(";"):
            (cursor or self.cursor).execute(statement.strip(), *[params])
        if commit:
            self.commit()

    def execute_and_return_result(self, sql, params=None, return_dict=False):
        if ";" in sql.strip()[:-1]:
            # Need to refactor this to properly deal with the single-statement constraint
            raise Exception("Invalid character ';' in statement!")
        if return_dict:
            raise Exception("return_dict = True not yet implemented for Snowflake!")
        cursor = self.cursor
        self.execute(sql=sql, params=params, commit=False, cursor=cursor)
        return [row for row in cursor]

    def close(self):
        if hasattr(self, "_cur"):
            try:
                # Uncommitted changes must be rolled back before closing cursor!
                self._cur.execute("ROLLBACK;")
            finally:
                self._cur.close()
                del self._cur
        if hasattr(self, "_snow_conn"):
            self._snow_conn.close()
            del self._snow_conn

    def __del__(self):
        self.close()



from ewah.constants import EWAHConstants as EC
from ewah.uploaders.snowflake import EWAHSnowflakeUploader
from ewah.uploaders.postgres import EWAHPostgresUploader
from ewah.uploaders.google_sheets import EWAHGSheetsUploader


def get_uploader(dwh_engine):
    try:
        return {
            EC.DWH_ENGINE_POSTGRES: EWAHPostgresUploader,
            EC.DWH_ENGINE_SNOWFLAKE: EWAHSnowflakeUploader,
            # DWH_ENGINE_BIGQUERY: bq_hook,
            EC.DWH_ENGINE_GS: EWAHGSheetsUploader,
        }[dwh_engine]
    except KeyError:
        raise Exception("Invalid Engine operator selected!")



from ewah.hooks.base import EWAHBaseHook

from tempfile import NamedTemporaryFile
from typing import Tuple, Optional, Dict, Any

import sshtunnel
import os

# The default timeout of 10s is too little if airflow runs many EL tasks in
# parallel -> if the SSH connection attempts are done in parallel, the timeout
# hits before the connection is established.
sshtunnel.TUNNEL_TIMEOUT = 30


class EWAHSSHHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "username": "login",
    }

    conn_name_attr = "ewah_ssh_conn_id"
    default_conn_name = "ewah_ssh_default"
    conn_type = "ewah_ssh"
    hook_name = "EWAH SSH Connection"

    @staticmethod
    def get_connection_form_widgets() -> Dict[str, Any]:
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from ewah.ewah_utils.widgets import EWAHTextAreaWidget

        return {
            "extra__ewah_ssh__private_key": StringField(
                "Private SSH Key (RSA Format)",
                widget=EWAHTextAreaWidget(rows=12),
            ),
            "extra__ewah_ssh__ssh_proxy_server": StringField(
                "SSH Proxy Server (optional)",
                widget=BS3TextFieldWidget(),
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        return {
            "hidden_fields": ["schema", "extra"],
            "relabeling": {
                "host": "SSH Host",
                "port": "SSH Port (defaults to 22)",
                "login": "SSH Username",
                "password": "SSH Password OR Private Key Password",
            },
        }

    def start_tunnel(
        self, remote_host: str, remote_port: int, tunnel_timeout: Optional[int] = 30
    ) -> Tuple[str, int]:
        """Starts the SSH tunnel with port forwarding. Returns the host and port tuple
        that can be used to connect to the remote.

        :param remote_host: Host of the remote that the tunnel should port forward to.
            Can be "localhost" e.g. if tunneling into a server that hosts a database.
        :param remote_port: Port that goes along with remote_host.
        :param tunnel_timeout: Optional timeout setting. Supply a higher number if the
            default (30s) is too low.
        :returns: Local bind address aka tuple of local_bind_host and local_bind_port.
            Calls to local_bind_host:local_bind_port will be forwarded to
            remote_host:remote_port via the SSH tunnel.
        """

        if not hasattr(self, "_ssh_tunnel_forwarder"):
            # Tunnel is not started yet - start it now!

            # Set a specific tunnel timeout if applicable
            if tunnel_timeout:
                old_timeout = sshtunnel.TUNNEL_TIMEOUT
                sshtunnel.TUNNEL_TIMEOUT = tunnel_timeout

            try:
                # Build kwargs dict for SSH Tunnel Forwarder
                if self.conn.ssh_proxy_server:
                    # Use the proxy SSH server as target
                    self._ssh_hook = EWAHBaseHook.get_hook_from_conn_id(
                        conn_id=self.conn.ssh_proxy_server,
                    )
                    kwargs = {
                        "ssh_address_or_host": self._ssh_hook.start_tunnel(
                            self.conn.host, self.conn.port or 22
                        ),
                        "remote_bind_address": (remote_host, remote_port),
                    }
                else:
                    kwargs = {
                        "ssh_address_or_host": (self.conn.host, self.conn.port or 22),
                        "remote_bind_address": (remote_host, remote_port),
                    }

                if self.conn.username:
                    kwargs["ssh_username"] = self.conn.username
                if self.conn.password:
                    kwargs["ssh_password"] = self.conn.password

                # Save private key in a temporary file, if applicable
                with NamedTemporaryFile() as keyfile:
                    if self.conn.private_key:
                        keyfile.write(self.conn.private_key.encode())
                        keyfile.flush()
                        kwargs["ssh_pkey"] = os.path.abspath(keyfile.name)
                    self.log.info(
                        "Opening SSH Tunnel to {0}:{1}...".format(
                            *kwargs["ssh_address_or_host"]
                        )
                    )
                    self._ssh_tunnel_forwarder = sshtunnel.SSHTunnelForwarder(**kwargs)
                    self._ssh_tunnel_forwarder.start()
            except:
                # Set package constant back to original setting, if applicable
                if tunnel_timeout:
                    sshtunnel.TUNNEL_TIMEOUT = old_timeout
                raise

        return ("localhost", self._ssh_tunnel_forwarder.local_bind_port)

    def stop_tunnel(self) -> None:
        """Close an open SSH tunnel, if it is indeed open."""
        if hasattr(self, "_ssh_tunnel_forwarder"):
            self.log.info(
                "Closing SSH tunnel to {0}!".format(
                    str(self._ssh_tunnel_forwarder._remote_binds)
                )
            )
            self._ssh_tunnel_forwarder.stop()
            del self._ssh_tunnel_forwarder
            if hasattr(self, "_ssh_hook"):
                self._ssh_hook.stop_tunnel()

    def __del__(self) -> None:
        self.stop_tunnel()
